"""AWS Secret Manager.

This module provides a secret server using the AWS system.
"""

import os
import logging
import typing
import json


try:
    import boto3
except Exception as problem:  # pylint: disable=broad-except
    logging.warning(
        'Unable to import boto3 because %s; aws secrets unavailable',
        str(problem))


from ox_secrets import settings
from ox_secrets.core import common


class AWSSecretServer(common.SecretServer):
    """Class to handle getting secrets from AWS.
    """

    @classmethod
    def load_secret_from_aws(
            cls, secret_id: str, profile_name: typing.Optional[
                str] = None, **kwargs) -> typing.Dict[str, str]:
        """Helper to load given secret_id from AWS.
        """
        storage = kwargs.pop('storage', os.path.splitext(
            secret_id)[-1]).lower().lstrip('.')
        if profile_name is None:
            profile_name = settings.OX_SECRETS_AWS_PROFILE_NAME
        logging.warning(
            'Connecting to boto3 for profile %s to read secrets for %s',
            profile_name, secret_id)
        service_name = kwargs.pop('service_name', 'secretsmanager')
        session = boto3.Session(profile_name=profile_name, **kwargs)
        client = session.client(service_name=service_name)
        if service_name == 'secretsmanager':
            get_secret_value_response = client.get_secret_value(
                SecretId=secret_id)
            secret_data = get_secret_value_response['SecretString']
        elif service_name == 'ssm':
            get_secret_value_response = client.get_parameter(Name=secret_id)
            secret_data = get_secret_value_response['Parameter']['Value']
        else:
            raise ValueError(f'Invalid service_name: {service_name}')

        if storage == 'json':
            secret_dict = json.loads(secret_data)
        else:
            if storage != 'raw':
                logging.warning('Intpereting unknown storage as raw.')
            secret_dict = {secret_id: secret_data}

        return secret_dict

    @classmethod
    def load_cache(cls, name: typing.Optional[str] = None,
                   category: typing.Optional[str] = None,
                   loader_params: typing.Optional[dict] = None):
        """Load secrets and cache them from back-end store.

        :param name:  String name of secret desired. If this is None, then
                      all secrets for given category are loaded.

        :param category:  String name of secret desired. Same as secret-id for
                          AWS command line interface.


        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  Fill cls._cache with data from secrets store.
                  Sub-classes must implement.

        """
        loader_params = loader_params if loader_params is not None else {}
        data = cls.load_secret_from_aws(secret_id=category, **loader_params)
        assert isinstance(data, dict)
        with cls._lock:
            cdict = cls._cache.get(category, {})
            if not cdict:
                cls._cache[category] = cdict
            cdict.update(data)

    @classmethod
    def store_secrets(cls, new_secret_dict: typing.Dict[str, str],
                      category: str, **storage_params):
        """Implement as required by parent class.
        """
        cls.store_secrets_to_aws(new_secret_dict, category, **storage_params)

    @classmethod
    def store_secrets_to_aws(
            cls, new_secret_dict: typing.Dict[str, str], category: str,
            profile_name: typing.Optional[str] = None, **kwargs):
        """Helper to store secrets to back-end file as required by parent.

        :param new_secret_dict:  Dictionary or name value pairs for secrets.

        :param category:  String category for secrets to store.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  Store the secrets in new_secret_dict to AWS.

        """
        logging.warning('Reading original secrets for "%s"', category)
        my_secrets = cls.load_secret_from_aws(
            category, profile_name, **kwargs)

        service_name = kwargs.pop('service_name', 'secretsmanager')
        session = boto3.Session(profile_name=profile_name, **kwargs)
        client = session.client(service_name=service_name)

        with cls._lock:
            my_secrets.update(new_secret_dict)
            secret_data = json.dumps(my_secrets)
            if service_name == 'secretsmanager':
                response = client.put_secret_value(
                    SecretId=category, SecretString=secret_data)
                logging.debug('Got response %s from AWS.', response)
            elif service_name == 'ssm':
                response = client.put_parameter(
                    Name=category, Value=secret_data, Type='String',
                    Overwrite=True)
            else:
                raise ValueError(f'Invalid service_name: {service_name}')
            cls._cache[category] = my_secrets



"""Environment variable server.

This module provides a secret server using environment variables
This is helpful for local testing and very simple operation.
"""

import typing
import logging
import os
import re

from ox_secrets.core import common


class EnvVarSecretServer(common.SecretServer):
    """Class to handle getting secrets from env variables.

Most of the other secret servers will check the environment for
a secret first by default. This class is mainly useful if
you *ONLY* want to get secrets from the environment variables.

The following illustrates example usage:

>>> import os
>>> os.environ['OX_SECRETS_SOMECATEGORY_EXAMPLE_NAME'] = 'example_1'
>>> os.environ['OX_SECRETS_WEB_EXAMPLE_URL'] = 'https://localhost/foo'
>>> from ox_secrets.core.evs import EnvVarSecretServer as secrets
>>> secrets.clear_cache()  # Clear it to make sure we start fresh
>>> secrets.get_secret('EXAMPLE_NAME', category='SOMECATEGORY')
'example_1'
>>> secrets.get_secret('EXAMPLE_URL', category='WEB')
'https://localhost/foo'
    """

    @classmethod
    def load_secrets_data(cls, prefix: str = 'OX_SECRETS'):
        """Load secrets file from environment variables.

        :param prefix='OX_SECRETS':  Prefix for environment variables to
                                     intrepret as secrets. Anything of
                                     the form {prefix}_{category}_{name} will
                                     be stored.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  Fill cls._cache with data from secrets vars.

        """
        logging.warning(
            'Reading secrets from env variables with prefix "%s"', prefix)
        my_re = re.compile(f'^{prefix}_(?P<category>[^_]+)_(?P<secname>.+)$')
        for var_name, value in os.environ.items():
            match = my_re.match(var_name)
            if match:
                category = match.group('category')
                secname = match.group('secname')
                logging.info('Matched %s as %s in category %s', secname,
                             category, secname)
                if category not in cls._cache:
                    cls._cache[category] = {}
                cls._cache[category][secname] = value

    @classmethod
    def load_cache(cls, name: typing.Optional[str] = None,
                   category: typing.Optional[str] = None,
                   loader_params: typing.Optional[dict] = None):
        "Implement loading cache from a file."
        if loader_params:
            logging.warning('Ignoring loader_params in %s', cls.__name__)
        with cls._lock:
            return cls.load_secrets_data()

    @classmethod
    def store_secrets(cls, new_secret_dict: typing.Dict[str, str],
                      category: str, **storage_params):
        """Implement as required by parent class.
        """
        return cls.store_secrets_to_env(
            new_secret_dict, category, **storage_params)

    @classmethod
    def store_secrets_to_env(cls, new_secret_dict: typing.Dict[str, str],
                             category: str, prefix: str = 'OX_SECRETS'):
        """Store secrets in environment variables.

        :param new_secret_dict:  Dictionary of secrets to store.

        :param category:str:     Category for storing secrets in env.

        :param prefix='OX_SECRETS':  Prefix for environment variables to
                                     intrepret as secrets. Anything of
                                     the form {prefix}_{category}_{name} will
                                     be stored.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  Store secrets in environment variables.  This is mainly
                  only useful for testing since the secrets are effectively
                  only stored in memory and get forgotten once the process
                  exits.
        """
        if new_secret_dict:
            logging.warning(
                'Storing to env will be forgotten when pricess exits.')
        with cls._lock:
            if category not in cls._cache:
                cls._cache[category] = {}
            for name, value in new_secret_dict.items():
                full_name = f'{prefix}_{category.upper()}_{name.upper()}'
                os.environ[full_name] = value
                cls._cache[category][name] = value



"""Common framework for all secret servers.

This module provides a common core for secret servers.
"""

import logging
import os
import threading
import typing
import re

from ox_secrets import settings


class SecretServer:
    """Class to handle gettting secrets.
    """

    _env_var_prefix = 'OX_SECRETS'
    _lock = threading.Lock()  # Used to lock access to _cache
    _cache = {}
    _data_fields = (
        ('name', 'Name of the secret.'),
        ('category', 'Category the secret is in.'),
        ('value', 'String value of the secret'),
        ('notes', 'Optional notes about the secret.')
        )

    @classmethod
    def load_cache(cls, name: typing.Optional[str] = None,
                   category: typing.Optional[str] = None,
                   loader_params: typing.Optional[dict] = None):
        """Load secrets and cache them from back-end store.

        :param name:  String name of secret desired. Some back-ends can load
                      the cache for all secrets at once while others require
                      the name and/or category. You should be able to pass
                      None for name to indicate loading everything for
                      the given category.

        :param category:  String name of secret desired. Some back-ends can
                          load the cache for all secrets at once while others
                          require the name and/or category.


        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  Fill cls._cache with data from secrets store.
                  Sub-classes must implement.

        *IMPORTANT:  Sub-classes should use `with cls._lock` inside when
                     overriding load_cache to be thread safe.
        """
        raise NotImplementedError

    @classmethod
    def clear_cache(cls):
        "Clear the cache (i.e., forget all secrets)."
        cls._cache = {}

    @classmethod
    def make_env_var_key(cls, name: str, category: str):
        """Create key to lookup environment variable for given name/category.
        """

        key = f'{cls._env_var_prefix}_{category}_{name}'.upper()
        return key

    @classmethod
    def secret_from_env(cls, name: str, category: str, allow_env: bool):
        """Try to extract secret from environment variable.

        :param name:  String name of secret.

        :param category:  String category.

        :param allow_env:  If False, return None.
        """
        if not allow_env:
            logging.debug('allow_env is false so not checking env')
            return None
        key = cls.make_env_var_key(name, category)
        result = os.environ.get(key, None)
        if result is not None:
            logging.info('Secret extracted from env var "%s"', key)
        return result

    @classmethod
    def get_secret(cls, name: str, category: str = 'root',
                   allow_env: bool = True,
                   allow_update: bool = True,
                   loader_params: typing.Optional[dict] = None,
                   service_name: typing.Optional[str] = None):
        """Lookup secret with given name and return it.

        :param name:     String name of secret to lookup.

        :param category='root':  Optional string category for secret

        :param allow_env=True:  Whether to allow checking env var first.

        :param allow_update=True:  Whether to loading cache if secret not
                                   found.

        :param loader_params: Optional dict of parameters which gets
                              passed to load_cache for back-end. This allows
                              a way to pass back-end specific info if desired.

        :param service_name: Optional string to add as
                             loader_params['service_name']. The
                             service_name is used by the AWS secret manager
                             and parameter store and nice to be able to
                             specify directly.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:   Value of secret.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:   Simple way to lookup secrets.

        """
        category = cls.maybe_replace_category(category)
        result = cls.secret_from_env(name, category, allow_env)
        if result is not None:
            return result
        with cls._lock:
            cdict = cls._cache.get(category, {})
            if name in cdict:
                return cdict[name]
            # Secret not found so clear cache to rebuild
            cls._cache[category] = None
            del cls._cache[category]
        cls._prepare_secrets_dict(category, allow_update,
                                  loader_params, service_name)

        with cls._lock:  # must lock since cdict may refernce cls._cdict
            cdict = cls._cache[category]
            return cdict[name]

    @classmethod
    def get_secret_dict(cls, category: str = 'root',
                        allow_update: bool = True,
                        loader_params: typing.Optional[dict] = None,
                        service_name: typing.Optional[str] = None):
        """Lookup secret with given name and return it.

        :param category='root':  Optional string category for secret

        :param allow_update=True:  Whether to loading cache if secret not
                                   found.

        :param loader_params: Optional dict of parameters which gets
                              passed to load_cache for back-end. This allows
                              a way to pass back-end specific info if desired.

        :param service_name: Optional string to add as
                             loader_params['service_name']. The service_name
                              is used by the AWS secret manager and parameter
                             store and nice to be able to specify directly.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:   Value of secret.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:   Simple way to lookup secrets.

        """
        category = cls.maybe_replace_category(category)
        cls._prepare_secrets_dict(category, allow_update, loader_params,
                                  service_name)
        with cls._lock:
            return dict(cls._cache[category])  # return shallow copy

    @classmethod
    def _prepare_secrets_dict(
            cls, category: str, allow_update: bool,
            loader_params: typing.Optional[dict] = None,
            service_name: typing.Optional[str] = None):
        """Help prepare secrets in cls._cache for get_secret and get_secret.

        :param category:  String category for secrets.

        :param allow_update=True:  Whether to loading cache if secret not
                                   found.

        :param loader_params: Optional dict of parameters which gets
                              passed to load_cache for back-end. This allows
                              a way to pass back-end specific info if desired.

        :param service_name: Optional string to add as
                             loader_params['service_name']. The service_name
                              is used by the AWS secret manager and parameter
                             store and nice to be able to specify directly.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:   This is a helper function to only be called by
                   get_secret and get_secret_dict. If cls._cache[category]
                   is non-empty, then this does nothign since we assume
                   we already have secrets loaded

        """

        if cls._cache.get(category):  # have something so no need to reload
            return
        if not allow_update:
            with cls._lock:
                cdict = cls._cache.get(category, None)
                if cdict is not None:   # did have something there
                    return              # so stop processing
                logging.error(          # otherwise indicate an error
                    'Unable to find category %s for secret manager class %s',
                    category, cls.__name__)
                raise KeyError(category)  # no data and no update allowed
        if service_name is not None:
            if loader_params is None:
                loader_params = {}
            loader_params['service_name'] = service_name
        cls.load_cache(name=None, category=category,
                       loader_params=loader_params)

    @staticmethod
    def maybe_replace_category(category):
        """Replace input based on OX_SECRETS_CATEGORY_REGEXP.

Meant to be called by get_secret.
        """
        if settings.OX_SECRETS_CATEGORY_REGEXP:
            return re.sub(settings.OX_SECRETS_CATEGORY_REGEXP,
                          settings.OX_SECRETS_CATEGORY_REPLACE, category)
        return category

    @classmethod
    def list_secret_names(cls, category: str) -> typing.List[str]:
        "Return list of secret names for given category."

        with cls._lock:  # get the lock to prevent modification while we look
            cdict = cls._cache.get(category, [])
            return list(cdict)

    @classmethod
    def store_secrets(cls, new_secret_dict: typing.Dict[str, str],
                      category: str, **storage_params):
        """Store secrets to back-end

        :param new_secret_dict:  Dictionary or name value pairs for secrets.

        :param category:  String category for secrets to store.

        :param **storage_params:   Optional storage parameters for back-end.
        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  Store the secrets in new_secret_dict.

        """
        raise NotImplementedError


class SecretInfo:
    """Class to hold information about a secret and convert to/from string.

The SecretInfo class is useful as a way to hold information about a
secret in a way which can be serialized to/from a string.

The following illustrates example usage:

>>> import tempfile, os                  # first do some boilerplate
>>> from ox_secrets.core import common   # imports of various things
>>> from ox_secrets import server
>>> tfile = tempfile.mktemp(suffix='.csv')  # setup a temp file
>>> open(tfile, 'wb').write(                # and write example secrets
... '''name,category,value,notes            # with an ususual encoding
... cool,thing,here,example'''.encode('UTF-32'))
352
>>> sstr = (  # create a secret serialized as a string
...   f'name=cool:category=thing:mode=fss:filename={tfile}:encoding=utf32')
>>> sec = common.SecretInfo.from_str(sstr)  # parse the string
>>> sec  # show pretty representation  doctest: +ELLPISIS
SecretInfo(name='cool',
  category='thing',
  mode='fss',
  filename='....csv',
  encoding='utf32')
>>> sstr == str(sec)  # verify we can convert back to string
True
>>> server.get_server(mode=sec.mode).get_secret(  # read secret
...     name=sec.name, category=sec.category, loader_params=sec.loader_params)
'here'
>>> os.remove(tfile)  # remove temp file

    """

    def __init__(self, name, category, mode=None, **kwargs):
        self.name = name
        self.category = category
        self.mode = mode
        self.loader_params = kwargs

    @classmethod
    def from_str(cls, my_str):
        """Convert a string representation of the secret to this class.

        :param my_str:   String representation of a secret.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:  Instance of class

        """
        kwargs = {}
        data = my_str.split(':')
        for item in data:
            key, value = item.split('=')
            kwargs[key.strip()] = value.strip()
        result = cls(**kwargs)
        return result

    def make_info(self):
        """Return list of name value pairs for data in self.
        """
        info = [('name', self.name), ('category', self.category)]
        if self.mode is not None:
            info.append(('mode', self.mode))
        for name, value in self.loader_params.items():
            info.append((name, value))
        return info

    def pretty(self):
        "Return pretty string reprsentation."
        info = self.make_info()
        arg_info = ",\n  ".join([f'{k}={repr(v)}' for k, v in info])
        return self.__class__.__name__ + f'({arg_info})'

    def __str__(self):
        info = self.make_info()
        return ':'.join([f'{k}={v}' for k, v in info])

    def __repr__(self):
        return self.pretty()



"""File system server.

This module provides a secret server using the file system.
This is helpful for local testing and very simple operation.
"""


import logging
import csv
import os
import typing
import json

from ox_secrets import settings
from ox_secrets.core import common


class FileSecretServer(common.SecretServer):
    """Class to handle getting secrets from file.
    """

    @classmethod
    def load_secrets_file(cls, filename=None, encoding='utf8',
                          file_type=None, default_category='root'):
        """Load secrets file from given filename.

        :param filename=None:    Optional filename for secrets. This
                                 defaults to ~/.ox_secrets.csv.
                                 It shall be a CSV file with header
                                 category,name,value,notes. You can also
                                 provide a .json file containing a dict
                                 of name value pairs which will all go
                                 in the default_category.

        :param encoding='utf8':  Default encoding for opening files.

        :param file_type:  Either .csv or .json or None if you want to
                           select based on file extension.

        :param default_category='root':  Category to use when cannot
                                         otherwise find it.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  Fill cls._cache with data from secrets file.

        """
        if filename is None:
            filename = os.environ.get(
                'OX_SECRETS_FILE', settings.OX_SECRETS_FILE)
        logging.warning('Opening secrets file "%s"', filename)
        with cls._lock, open(filename, 'r', encoding=encoding) as sfd:
            if file_type is None:
                file_type = os.path.splitext(filename)[-1].lower()
            if file_type == '.raw':
                contents = [{'name': default_category, 'value': sfd.read()}]
            elif file_type == '.csv':
                contents = list(csv.DictReader(sfd))
            elif file_type == '.json':
                contents = [{'name': k, 'value': v}
                            for k, v in json.load(sfd).items()]
            else:
                raise ValueError(
                    f'Cannot handle secrets file_type={file_type}')
            for line in contents:
                line_category = line.get('category', default_category)
                if line_category not in cls._cache:
                    cls._cache[line_category] = {}
                cls._cache[line_category][line['name']] = line[
                    'value']

    @classmethod
    def load_cache(cls, name: typing.Optional[str] = None,
                   category: typing.Optional[str] = None,
                   loader_params: typing.Optional[dict] = None):
        "Implement loading cache from a file."
        loader_params = loader_params if loader_params is not None else {}
        return cls.load_secrets_file(default_category=category,
                                     **loader_params)

    @classmethod
    def store_secrets(cls, new_secret_dict: typing.Dict[str, str],
                      category: str, **storage_params):
        """Implement as required by parent class.
        """
        cls.store_secrets_to_file(new_secret_dict, category, **storage_params)

    @classmethod
    def store_secrets_to_file(
            cls, new_secret_dict: typing.Dict[str, str], category: str,
            filename: typing.Optional[str] = None,
            encoding: str = 'utf8'):
        """Helper to store secrets to back-end file as required by parent.

        :param new_secret_dict:  Dictionary or name value pairs for secrets.

        :param category:  String category for secrets to store.

        :param filename: Path to store secrets to.

        :param encoding: Encoding for file.
        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  Store the secrets in new_secret_dict to filename.

        """
        if filename is None:
            filename = os.environ.get(
                'OX_SECRETS_FILE', settings.OX_SECRETS_FILE)
        logging.warning('Opening secrets file "%s"', filename)
        data = []
        with cls._lock:
            if not os.path.exists(filename):
                logging.warning('Creating empty file for %s.', filename)
                with open(filename, 'a', encoding='utf8') as sfd:
                    pass  # this just creates an empty file if necessary
            with open(filename, 'r', encoding=encoding) as sfd:
                reader = csv.DictReader(sfd)
                for item in reader:
                    if item['name'] in new_secret_dict and item[
                            'category'] == category:
                        logging.info('Replacing old value for %s/%s',
                                     item['name'], item['category'])
                    else:
                        data.append(item)
            with open(filename, 'w', encoding=encoding) as sfd:
                writer = csv.DictWriter(sfd, fieldnames=[
                    pair[0] for pair in cls._data_fields])
                writer.writeheader()
                for line in data:
                    writer.writerow(line)
                for name, value in new_secret_dict.items():
                    writer.writerow({
                        'name': name, 'category': category,
                        'value': value, 'notes': 'via store_secrets'})
                    if category in cls._cache:
                        cls._cache[category][name] = value



"""Simple secret server.

The ox_secrets package provides a simple secret server with various
back-ends. The following illustrates example usage.

First we setup an example secrets file:

>>> import os, tempfile, csv
>>> fn = tempfile.mktemp(suffix='_ox_secrets.csv')
>>> writer = csv.writer(open(fn, 'w')).writerows([
... ['name', 'category', 'value', 'notes'],
... ['example_name', 'root', 'super_secret', 'example secret'],
... ['example_pw', 'prod/data', 'super_secret_pw', 'example secret_pw'],
... ['example_pw', 'test/data', 'unsecret_test_pw', 'example secret test pw'],
... ['example_pw', 'alt', 'alt_unsecret_test_pw', 'alt secret test pw']])
>>> print(open(fn).read().strip())
name,category,value,notes
example_name,root,super_secret,example secret
example_pw,prod/data,super_secret_pw,example secret_pw
example_pw,test/data,unsecret_test_pw,example secret test pw
example_pw,alt,alt_unsecret_test_pw,alt secret test pw

Now that we have a secrets file, we can read some secrets:

>>> from ox_secrets import settings, server as oss
>>> oss.settings.OX_SECRETS_FILE = fn # default is ~/.ox_secrets.csv
>>> oss.forget_secrets()  # Clear it to make sure we start fresh
>>> oss.get_secret('example_name')
'super_secret'

We can also get a dictionary of all the secrets for a given category:

>>> oss.get_secret_dict(category='test/data')
{'example_pw': 'unsecret_test_pw'}

Sometimes it is nice to be able to just pass a dictionary of
credential information to get_secret:

>>> creds = {'name': 'example_name', 'category': 'root', 'server': 'fss'}
>>> oss.get_secret(**creds)
'super_secret'

You can also puts secrets into the environment variables:

>>> os.environ['OX_SECRETS_ROOT_EXAMPLE_NAME'] = 'other'
>>> oss.get_secret('example_name')
'other'

You can use the OX_SECRETS_CATEGORY_REGEXP and
the OX_SECRETS_CATEGORY_REPLACE either in the settings file
or environment variables (before starting python) to automatically
switch from production to testing secrets:

>>> oss.get_secret('example_pw', 'prod/data')
'super_secret_pw'
>>> oss.forget_secrets()  # Clear it to make sure we start fresh
>>> oss.settings.OX_SECRETS_CATEGORY_REGEXP = '^prod/'
>>> oss.settings.OX_SECRETS_CATEGORY_REPLACE = 'test/'
>>> oss.get_secret('example_pw', 'prod/data')
'unsecret_test_pw'


If desired, you can also store secrets (assuming
you have appropriate permissions):

>>> oss.store_secrets({'example_pw': 'foobar'}, category='alt')
>>> oss.get_secret('example_pw', category='alt')
'foobar'

Now cleanup

>>> os.remove(fn)


If you have an account with the appropriate permissions (e.g.,
you may need to set the AWS_PROFILE environment variable to
to such an account), you can also store secrets and parameters
to aws.

For example, you could do something like:

    oss.get_server(mode='aws').store_secrets(
        {'test_storage':'foobar'}, category=AWS_SECRET_ID)

to store a secret to the existing secret with secret ID
`AWS_SECRET_ID` on Amazon Web Services. You could also proide the
`service_name='ssm'` argument if you wanted to use the parameter store
instead of the secret store via something like:

    oss.get_server(mode='aws').store_secrets(
        {'test_storage':'foobar', category=AWS_PARAM_NAME,
        service_name='ssm')


"""

VERSION = '0.5.5'
__version__ = VERSION



"""Settings for how to use ox_secrets
"""

import os

# Specifies which type of secret server you
# want to use. Can change this at run-time
# or override with environment variable OX_SECRETS_MODE
OX_SECRETS_MODE = os.environ.get('OX_SECRETS_MODE', 'FILE')

# Specifies default location of secrets file for server from
# file system in fss.py. Default is ~/.ox_secrets. You can
# override either with environment variable OX_SECRETS_FILE
# or by changing below at run-time.
OX_SECRETS_FILE = os.path.join(os.environ.get(
    'OX_SECRETS_FILE', os.environ.get('HOME', '/')), '.ox_secrets.csv')


# Used for the aws secrets manager. Specifies the profile name in your ~/.aws.
OX_SECRETS_AWS_PROFILE_NAME = None

# If you use '^prod/' for OX_SECRETS_CATEGORY_REGEXP and
# use 'test/' for OX_SECRETS_CATEGORY_REPLACE then you will
# automatically have the all categories starting with 'prod/' replaced
# with 'test/' in calls to get_secret.
OX_SECRETS_CATEGORY_REGEXP = os.environ.get(
    'OX_SECRETS_CATEGORY_REGEXP', None)
OX_SECRETS_CATEGORY_REPLACE = os.environ.get(
    'OX_SECRETS_CATEGORY_REPLACE', None)



"""Simple Command Line Interface to get secrets.
"""

import logging
import argparse

from ox_secrets import server


def prep_parser():
    """Prepare parser for command line.
    """
    parser = argparse.ArgumentParser(
        prog='ox_secrets',
        description='Command Line Interface to lookup secrets.')
    parser.add_argument(
        'name', help='Name of the secret to lookup.')
    parser.add_argument(
        '--raise-error', action='store_true',
        help='Force raising exception to see stack trace.')
    parser.add_argument('--loglevel', default='ERROR', help=(
        'Log level to use (e.g., DEBUG, INFO, WARNING, ERROR).'))
    parser.add_argument(
        '--category', default='root', help=(
            'Category of the secret to lookup (default is "root")'))
    parser.add_argument(
        '--mode', default=None, help=(
            'Mode to use in finding secret server.'))
    parser.add_argument(
        '--loader', action='append', help=(
            'Loader params (e.g., service_name/ssm)'
            ' to use AWS parameter store'
            ))

    return parser


def main():
    """Main implementation of command line.
    """
    parser = prep_parser()
    args = parser.parse_args()
    result = 'unknown error'
    try:
        logging.getLogger('').setLevel(getattr(logging, args.loglevel))
        my_server = server.get_server(mode=args.mode)
        loader_params = {}
        for item in (args.loader or []):
            param_name, *param_value = item.split('/')
            param_value = '/'.join(param_value)
            loader_params[param_name] = param_value
        result = my_server.get_secret(
            name=args.name, category=args.category,
            loader_params=loader_params)
    except Exception as problem:  # pylint: disable=broad-except
        if args.raise_error:
            raise
        msg = (f'Error in CLI: {problem}.\n'
               'Pass --raise-error to force raising to see stack trace.')
        logging.error(msg)
        result = 'Error in CLI.'

    print(result)


if __name__ == '__main__':
    main()



"""High-level file to work with secret servers.
"""

import os
import typing

from ox_secrets import settings
from ox_secrets.core import common, fss, evs


def get_server(mode: typing.Optional[str] = None) -> common.SecretServer:
    """Get the default secret server.

    :param mode:  Optional string mode specifying secret server to get. If
                  None, then we try OX_SECRETS_MODE from env and then
                  settings.OX_SECRETS_MODE.

    """
    if not mode:
        mode = os.environ.get('OX_SECRETS_MODE', settings.OX_SECRETS_MODE)
    mode = mode.lower()
    if mode in ['fss', 'file']:
        return fss.FileSecretServer
    if mode in ['evs', 'env']:
        return evs.EnvVarSecretServer
    if mode == 'aws':
        # delayed import so boto3 does not need to be installed if
        # the aws secret server is never used
        from ox_secrets.core import aws  \
            # pylint:disable=import-outside-toplevel
        return aws.AWSSecretServer
    raise ValueError(f'Invalid secret server mode "{mode}"')


def get_secret(name: str, category: str = 'root',
               server: typing.Optional[
                   typing.Union[str, common.SecretServer]] = None) -> str:
    """Return desired secret.

    :param name: str: Name of secret to get

    :param category: str ='root': Category of secret.

    :param server=None:  Optional server to use (either the class or
                         string name for it such as 'fss' or 'aws'. If
                         not provided, then we call get_server() to
                         get it.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    :return:  Desired secret.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    PURPOSE:  This is a convenience function and meant to be the main
              function that most applications need to call.

    """
    if server is None or isinstance(server, str):
        server = get_server(mode=server)
    result = server.get_secret(name, category)
    return result


def get_secret_dict(
        mode: typing.Optional[str] = None, category: str = 'root',
        allow_update: bool = True,
        loader_params: typing.Optional[dict] = None,
        service_name: typing.Optional[str] = None) -> typing.Dict[
            str, typing.Any]:
    """Lookup secret with given name and return it.

    :param mode=None:  Optional string mode for calling get_server.

    :param category='root':  Optional string category for secret

    :param allow_update=True:  Whether to loading cache if secret not
                               found.

    :param loader_params: Optional dict of parameters which gets
                          passed to load_cache for back-end. This allows
                          a way to pass back-end specific info if desired.

    :param service_name: Optional string to add as
                         loader_params['service_name']. The service_name
                          is used by the AWS secret manager and parameter
                         store and nice to be able to specify directly.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    :return:   Dictionary of secrets for given category.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    PURPOSE:   Simple way to lookup dictionary of secrets.

    """
    server = get_server(mode=mode)
    return server.get_secret_dict(
        category=category, allow_update=allow_update,
        loader_params=loader_params, service_name=service_name)


def forget_secrets(server: common.SecretServer = None) -> str:
    """Forget all secrets.
    """
    if server is None:
        server = get_server()
    server.clear_cache()


def store_secrets(new_secret_dict: typing.Dict[str, str],
                  category: str, server: typing.Optional[
                      typing.Union[str, common.SecretServer]] = None,
                  **storage_params):
    """Store secrets to back-end

    :param new_secret_dict:  Dictionary or name value pairs for secrets.

    :param category:  String category for secrets to store.

    :param server:  Secret server to use.

    :param **storage_params:   Optional storage parameters for back-end.
    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    PURPOSE:  Store the secrets in new_secret_dict.

    """
    if server is None:
        server = get_server()
    server.store_secrets(new_secret_dict, category, **storage_params)
